import fiona
import pytest
from fiona.drvsupport import driver_mode_mingdal, _driver_supports_mode
from fiona.errors import DriverError
from tests.conftest import get_temp_filename


@pytest.mark.parametrize(
    "driver",
    [
        driver
        for driver in driver_mode_mingdal["w"].keys()
        if _driver_supports_mode(driver, "w")
    ],
)
def test_write_getextent(tmpdir, driver, testdata_generator):
    """Test if a call to OGR_L_GetExtent has side effects for writing."""
    schema, crs, records1, records2, _ = testdata_generator(
        driver, range(0, 10), range(10, 20)
    )
    path = str(tmpdir.join(get_temp_filename(driver)))
    positions = {int(r['properties']['position']) for r in records1 + records2}

    with fiona.open(
        path,
        "w",
        driver=driver,
        crs=crs,
        schema=schema,
    ) as c:
        c.writerecords(records1)

        # Call to OGR_L_GetExtent
        try:
            c.bounds
        except DriverError:
            pass

        c.writerecords(records2)

    with fiona.open(path) as c:
        data = {int(f['properties']['position']) for f in c}
        assert len(positions) == len(data)
        for p in positions:
            assert p in data


@pytest.mark.parametrize(
    "driver",
    [
        driver
        for driver in driver_mode_mingdal["w"].keys()
        if _driver_supports_mode(driver, "w")
    ],
)
def test_read_getextent(tmpdir, driver, testdata_generator):
    """Test if a call to OGR_L_GetExtent has side effects for reading."""
    schema, crs, records1, records2, _ = testdata_generator(
        driver, range(0, 10), range(10, 20)
    )
    path = str(tmpdir.join(get_temp_filename(driver)))
    positions = {int(r['properties']['position']) for r in records1 + records2}

    with fiona.open(
        path,
        "w",
        driver=driver,
        crs=crs,
        schema=schema,
    ) as c:
        c.writerecords(records1)
        c.writerecords(records2)

    with fiona.open(path) as c:
        data = set()
        for _ in range(len(records1)):
            f = next(c)
            data.add(int(f['properties']['position']))

        # Call to OGR_L_GetExtent
        try:
            c.bounds
        except DriverError:
            pass

        for _ in range(len(records1)):
            f = next(c)
            data.add(int(f['properties']['position']))
        assert len(positions) == len(data)
        for p in positions:
            assert p in data


@pytest.mark.parametrize(
    "driver",
    [
        driver
        for driver in driver_mode_mingdal["w"].keys()
        if _driver_supports_mode(driver, "w")
    ],
)
def test_write_getfeaturecount(tmpdir, driver, testdata_generator):
    """Test if a call to OGR_L_GetFeatureCount has side effects for writing."""
    schema, crs, records1, records2, _ = testdata_generator(
        driver, range(0, 10), range(10, 20)
    )
    path = str(tmpdir.join(get_temp_filename(driver)))
    positions = {int(r['properties']['position']) for r in records1 + records2}

    with fiona.open(
        path,
        "w",
        driver=driver,
        crs=crs,
        schema=schema,
    ) as c:
        c.writerecords(records1)

        # Call to OGR_L_GetFeatureCount
        try:
            assert len(c) == len(records1)
        except TypeError:
            pass
        c.writerecords(records2)

    with fiona.open(path) as c:
        data = {int(f['properties']['position']) for f in c}
        assert len(positions) == len(data)
        for p in positions:
            assert p in data


@pytest.mark.parametrize(
    "driver",
    [
        driver
        for driver in driver_mode_mingdal["w"].keys()
        if _driver_supports_mode(driver, "w")
    ],
)
def test_read_getfeaturecount(tmpdir, driver, testdata_generator):
    """Test if a call to OGR_L_GetFeatureCount has side effects for reading."""
    schema, crs, records1, records2, _ = testdata_generator(
        driver, range(0, 10), range(10, 20)
    )
    path = str(tmpdir.join(get_temp_filename(driver)))
    positions = {int(r['properties']['position']) for r in records1 + records2}

    with fiona.open(
        path,
        "w",
        driver=driver,
        crs=crs,
        schema=schema,
    ) as c:
        c.writerecords(records1)
        c.writerecords(records2)

    with fiona.open(path) as c:
        data = set()
        for _ in range(len(records1)):
            f = next(c)
            data.add(int(f['properties']['position']))

        # Call to OGR_L_GetFeatureCount
        try:
            assert len(data) == len(records1)
        except TypeError:
            pass

        for _ in range(len(records1)):
            f = next(c)
            data.add(int(f['properties']['position']))

        try:
            assert len(data) == len(records1 + records2)
        except TypeError:
            pass

        assert len(positions) == len(data)
        for p in positions:
            assert p in data



"""fio-dump"""

from functools import partial
import json
import logging

import click
import cligj

import fiona
from fiona.fio import helpers, options, with_context_env
from fiona.model import Feature, ObjectEncoder
from fiona.transform import transform_geom


@click.command(short_help="Dump a dataset to GeoJSON.")
@click.argument('input', required=True)
@click.option('--layer', metavar="INDEX|NAME", callback=options.cb_layer,
              help="Print information about a specific layer.  The first "
                   "layer is used by default.  Layers use zero-based "
                   "numbering when accessed by index.")
@click.option('--encoding', help="Specify encoding of the input file.")
@cligj.precision_opt
@cligj.indent_opt
@cligj.compact_opt
@click.option('--record-buffered/--no-record-buffered', default=False,
              help="Economical buffering of writes at record, not collection "
                   "(default), level.")
@click.option('--ignore-errors/--no-ignore-errors', default=False,
              help="log errors but do not stop serialization.")
@click.option('--with-ld-context/--without-ld-context', default=False,
              help="add a JSON-LD context to JSON output.")
@click.option('--add-ld-context-item', multiple=True,
              help="map a term to a URI and add it to the output's JSON LD "
                   "context.")
@click.pass_context
@with_context_env
def dump(ctx, input, encoding, precision, indent, compact, record_buffered,
         ignore_errors, with_ld_context, add_ld_context_item, layer):

    """Dump a dataset either as a GeoJSON feature collection (the default)
    or a sequence of GeoJSON features."""

    logger = logging.getLogger(__name__)
    sink = click.get_text_stream('stdout')

    dump_kwds = {'sort_keys': True}
    if indent:
        dump_kwds['indent'] = indent
    if compact:
        dump_kwds['separators'] = (',', ':')
    item_sep = compact and ',' or ', '

    open_kwds = {}
    if encoding:
        open_kwds['encoding'] = encoding
    if layer:
        open_kwds['layer'] = layer

    def transformer(crs, feat):
        tg = partial(transform_geom, crs, 'EPSG:4326',
                     antimeridian_cutting=True, precision=precision)
        return Feature(id=feat.id, properties=feat.properties, geometry=tg(feat.geometry))

    with fiona.open(input, **open_kwds) as source:
        meta = source.meta
        meta["fields"] = dict(source.schema["properties"].items())

        if record_buffered:
            # Buffer GeoJSON data at the feature level for smaller
            # memory footprint.
            indented = bool(indent)
            rec_indent = "\n" + " " * (2 * (indent or 0))

            collection = {
                "type": "FeatureCollection",
                "fiona:schema": meta["schema"],
                "fiona:crs": meta["crs"],
                "features": [],
            }
            if with_ld_context:
                collection["@context"] = helpers.make_ld_context(add_ld_context_item)

            head, tail = json.dumps(collection, **dump_kwds).split("[]")

            sink.write(head)
            sink.write("[")

            itr = iter(source)

            # Try the first record.
            try:
                i, first = 0, next(itr)
                first = transformer(first)
                if with_ld_context:
                    first = helpers.id_record(first)
                if indented:
                    sink.write(rec_indent)
                sink.write(
                    json.dumps(first, cls=ObjectEncoder, **dump_kwds).replace(
                        "\n", rec_indent
                    )
                )
            except StopIteration:
                pass
            except Exception as exc:
                # Ignoring errors is *not* the default.
                if ignore_errors:
                    logger.error(
                        "failed to serialize file record %d (%s), " "continuing", i, exc
                    )
                else:
                    # Log error and close up the GeoJSON, leaving it
                    # more or less valid no matter what happens above.
                    logger.critical(
                        "failed to serialize file record %d (%s), " "quiting", i, exc
                    )
                    sink.write("]")
                    sink.write(tail)
                    if indented:
                        sink.write("\n")
                    raise

            # Because trailing commas aren't valid in JSON arrays
            # we'll write the item separator before each of the
            # remaining features.
            for i, rec in enumerate(itr, 1):
                rec = transformer(rec)
                try:
                    if with_ld_context:
                        rec = helpers.id_record(rec)
                    if indented:
                        sink.write(rec_indent)
                    sink.write(item_sep)
                    sink.write(
                        json.dumps(rec, cls=ObjectEncoder, **dump_kwds).replace(
                            "\n", rec_indent
                        )
                    )
                except Exception as exc:
                    if ignore_errors:
                        logger.error(
                            "failed to serialize file record %d (%s), "
                            "continuing",
                            i, exc)
                    else:
                        logger.critical(
                            "failed to serialize file record %d (%s), "
                            "quiting",
                            i, exc)
                        sink.write("]")
                        sink.write(tail)
                        if indented:
                            sink.write("\n")
                        raise

            # Close up the GeoJSON after writing all features.
            sink.write("]")
            sink.write(tail)
            if indented:
                sink.write("\n")

        else:
            # Buffer GeoJSON data at the collection level. The default.
            collection = {
                "type": "FeatureCollection",
                "fiona:schema": meta["schema"],
                "fiona:crs": meta["crs"].to_string(),
            }
            if with_ld_context:
                collection["@context"] = helpers.make_ld_context(add_ld_context_item)
                collection["features"] = [
                    helpers.id_record(transformer(rec)) for rec in source
                ]
            else:
                collection["features"] = [
                    transformer(source.crs, rec) for rec in source
                ]
            json.dump(collection, sink, cls=ObjectEncoder, **dump_kwds)



"""$ fio distrib"""

import json

import click
import cligj

from fiona.fio import helpers, with_context_env
from fiona.model import ObjectEncoder


@click.command()
@cligj.use_rs_opt
@click.pass_context
@with_context_env
def distrib(ctx, use_rs):
    """Distribute features from a collection.

    Print the features of GeoJSON objects read from stdin.

    """
    stdin = click.get_text_stream('stdin')
    source = helpers.obj_gen(stdin)

    for i, obj in enumerate(source):
        obj_id = obj.get("id", "collection:" + str(i))
        features = obj.get("features") or [obj]
        for j, feat in enumerate(features):
            if obj.get("type") == "FeatureCollection":
                feat["parent"] = obj_id
            feat_id = feat.get("id", "feature:" + str(i))
            feat["id"] = feat_id
            if use_rs:
                click.echo("\x1e", nl=False)
            click.echo(json.dumps(feat, cls=ObjectEncoder))



"""Tests of features related to GDAL RFC 64

See https://trac.osgeo.org/gdal/wiki/rfc64_triangle_polyhedralsurface_tin.
"""

import fiona
from fiona.model import Geometry


def _test_tin(geometry: Geometry) -> None:
    """Test if TIN (((0 0 0, 0 0 1, 0 1 0, 0 0 0)), ((0 0 0, 0 1 0, 1 1 0, 0 0 0)))
    is correctly converted to MultiPolygon.
    """
    assert geometry["type"] == "MultiPolygon"
    assert geometry["coordinates"] == [
        [[(0.0, 0.0, 0.0), (0.0, 0.0, 1.0), (0.0, 1.0, 0.0), (0.0, 0.0, 0.0)]],
        [[(0.0, 0.0, 0.0), (0.0, 1.0, 0.0), (1.0, 1.0, 0.0), (0.0, 0.0, 0.0)]],
    ]


def _test_triangle(geometry: Geometry) -> None:
    """Test if TRIANGLE((0 0 0,0 1 0,1 1 0,0 0 0))
    is correctly converted to MultiPolygon."""
    assert geometry["type"] == "Polygon"
    assert geometry["coordinates"] == [
        [(0.0, 0.0, 0.0), (0.0, 1.0, 0.0), (1.0, 1.0, 0.0), (0.0, 0.0, 0.0)]
    ]


def test_tin_shp(path_test_tin_shp):
    """Convert TIN to MultiPolygon"""
    with fiona.open(path_test_tin_shp) as col:
        assert col.schema["geometry"] == "Unknown"
        features = list(col)
        assert len(features) == 1
        _test_tin(features[0]["geometry"])


def test_tin_csv(path_test_tin_csv):
    """Convert TIN to MultiPolygon and Triangle to Polygon"""
    with fiona.open(path_test_tin_csv) as col:
        assert col.schema["geometry"] == "Unknown"

        feature1 = next(col)
        _test_tin(feature1["geometry"])

        feature2 = next(col)
        _test_triangle(feature2["geometry"])

        feature3 = next(col)
        assert feature3["geometry"]["type"] == "GeometryCollection"
        assert len(feature3["geometry"]["geometries"]) == 2

        _test_tin(feature3["geometry"]["geometries"][0])
        _test_triangle(feature3["geometry"]["geometries"][1])



"""$ fio insp"""


import code
import sys

import click

import fiona
from fiona.fio import with_context_env


@click.command(short_help="Open a dataset and start an interpreter.")
@click.argument('src_path', required=True)
@click.option('--ipython', 'interpreter', flag_value='ipython',
              help="Use IPython as interpreter.")
@click.pass_context
@with_context_env
def insp(ctx, src_path, interpreter):
    """Open a collection within an interactive interpreter."""
    banner = (
        "Fiona %s Interactive Inspector (Python %s)\n"
        'Type "src.schema", "next(src)", or "help(src)" '
        "for more information."
        % (fiona.__version__, ".".join(map(str, sys.version_info[:3])))
    )

    with fiona.open(src_path) as src:
        scope = locals()
        if not interpreter:
            code.interact(banner, local=scope)
        elif interpreter == "ipython":
            import IPython

            IPython.InteractiveShell.banner1 = banner
            IPython.start_ipython(argv=[], user_ns=scope)
        else:
            raise click.ClickException(
                "Interpreter {} is unsupported or missing "
                "dependencies".format(interpreter)
            )



from collections import OrderedDict
import os
import tempfile

import pytest

import fiona
from fiona.drvsupport import driver_mode_mingdal
from fiona.env import GDALVersion
from fiona.errors import SchemaError, UnsupportedGeometryTypeError, DriverSupportError
from fiona.model import Feature
from fiona.schema import FIELD_TYPES, normalize_field_type

from .conftest import get_temp_filename
from .conftest import requires_only_gdal1, requires_gdal2


def test_schema_ordering_items(tmpdir):
    name = str(tmpdir.join("test_scheme.shp"))
    items = [("title", "str:80"), ("date", "date")]
    with fiona.open(
        name,
        "w",
        driver="ESRI Shapefile",
        schema={"geometry": "LineString", "properties": items},
    ) as c:
        assert list(c.schema["properties"].items()) == items
    with fiona.open(name) as c:
        assert list(c.schema["properties"].items()) == items


def test_shapefile_schema(tmpdir):
    name = str(tmpdir.join("test_schema.shp"))
    items = sorted(
        {
            "AWATER10": "float",
            "CLASSFP10": "str",
            "ZipCodeType": "str",
            "EstimatedPopulation": "float",
            "LocationType": "str",
            "ALAND10": "float",
            "TotalWages": "float",
            "FUNCSTAT10": "str",
            "Long": "float",
            "City": "str",
            "TaxReturnsFiled": "float",
            "State": "str",
            "Location": "str",
            "GSrchCnt": "float",
            "INTPTLAT10": "str",
            "Lat": "float",
            "MTFCC10": "str",
            "Decommisioned": "str",
            "GEOID10": "str",
            "INTPTLON10": "str",
        }.items()
    )
    with fiona.open(
        name,
        "w",
        driver="ESRI Shapefile",
        schema={"geometry": "Polygon", "properties": items},
    ) as c:
        assert list(c.schema["properties"].items()) == items
        c.write(
            Feature.from_dict(
                **{
                    "geometry": {
                        "coordinates": [
                            [
                                (-117.882442, 33.783633),
                                (-117.882284, 33.783817),
                                (-117.863348, 33.760016),
                                (-117.863478, 33.760016),
                                (-117.863869, 33.760017),
                                (-117.864, 33.760017999999995),
                                (-117.864239, 33.760019),
                                (-117.876608, 33.755769),
                                (-117.882886, 33.783114),
                                (-117.882688, 33.783345),
                                (-117.882639, 33.783401999999995),
                                (-117.88259, 33.78346),
                                (-117.882442, 33.783633),
                            ]
                        ],
                        "type": "Polygon",
                    },
                    "id": "1",
                    "properties": {
                        "ALAND10": 8819240.0,
                        "AWATER10": 309767.0,
                        "CLASSFP10": "B5",
                        "City": "SANTA ANA",
                        "Decommisioned": False,
                        "EstimatedPopulation": 27773.0,
                        "FUNCSTAT10": "S",
                        "GEOID10": "92706",
                        "GSrchCnt": 0.0,
                        "INTPTLAT10": "+33.7653010",
                        "INTPTLON10": "-117.8819759",
                        "Lat": 33.759999999999998,
                        "Location": "NA-US-CA-SANTA ANA",
                        "LocationType": "PRIMARY",
                        "Long": -117.88,
                        "MTFCC10": "G6350",
                        "State": "CA",
                        "TaxReturnsFiled": 14635.0,
                        "TotalWages": 521280485.0,
                        "ZipCodeType": "STANDARD",
                    },
                    "type": "Feature",
                }
            )
        )
        assert len(c) == 1
    with fiona.open(name) as c:
        assert list(c.schema["properties"].items()) == sorted(
            [
                ("AWATER10", "float:24.15"),
                ("CLASSFP10", "str:80"),
                ("ZipCodeTyp", "str:80"),
                ("EstimatedP", "float:24.15"),
                ("LocationTy", "str:80"),
                ("ALAND10", "float:24.15"),
                ("INTPTLAT10", "str:80"),
                ("FUNCSTAT10", "str:80"),
                ("Long", "float:24.15"),
                ("City", "str:80"),
                ("TaxReturns", "float:24.15"),
                ("State", "str:80"),
                ("Location", "str:80"),
                ("GSrchCnt", "float:24.15"),
                ("TotalWages", "float:24.15"),
                ("Lat", "float:24.15"),
                ("MTFCC10", "str:80"),
                ("INTPTLON10", "str:80"),
                ("GEOID10", "str:80"),
                ("Decommisio", "str:80"),
            ]
        )
        f = next(iter(c))
        assert f.properties["EstimatedP"] == 27773.0


def test_field_truncation_issue177(tmpdir):
    name = str(tmpdir.join("output.shp"))

    kwargs = {
        "driver": "ESRI Shapefile",
        "crs": "EPSG:4326",
        "schema": {"geometry": "Point", "properties": [("a_fieldname", "float")]},
    }

    with fiona.open(name, "w", **kwargs) as dst:
        rec = {}
        rec["geometry"] = {"type": "Point", "coordinates": (0, 0)}
        rec["properties"] = {"a_fieldname": 3.0}
        dst.write(Feature.from_dict(**rec))

    with fiona.open(name) as src:
        first = next(iter(src))
        assert first.geometry.type == "Point"
        assert first.geometry.coordinates == (0, 0)
        assert first.properties["a_fieldnam"] == 3.0


def test_unsupported_geometry_type():
    tmpdir = tempfile.mkdtemp()
    tmpfile = os.path.join(tmpdir, "test-test-geom.shp")

    profile = {
        "driver": "ESRI Shapefile",
        "schema": {"geometry": "BOGUS", "properties": {}},
    }

    with pytest.raises(UnsupportedGeometryTypeError):
        fiona.open(tmpfile, "w", **profile)


@pytest.mark.parametrize("x", list(range(1, 10)))
def test_normalize_int32(x):
    assert normalize_field_type(f"int:{x}") == "int32"


@requires_gdal2
@pytest.mark.parametrize("x", list(range(10, 20)))
def test_normalize_int64(x):
    assert normalize_field_type(f"int:{x}") == "int64"


@pytest.mark.parametrize("x", list(range(0, 20)))
def test_normalize_str(x):
    assert normalize_field_type(f"str:{x}") == "str"


def test_normalize_bool():
    assert normalize_field_type("bool") == "bool"


def test_normalize_float():
    assert normalize_field_type("float:25.8") == "float"


def generate_field_types():
    """
    Produce a unique set of field types in a consistent order.

    This ensures that tests are able to run in parallel.
    """
    types = set(FIELD_TYPES)
    types.remove(None)
    return list(sorted(types)) + [None]


@pytest.mark.parametrize("x", generate_field_types())
def test_normalize_std(x):
    assert normalize_field_type(x) == x


def test_normalize_error():
    with pytest.raises(SchemaError):
        assert normalize_field_type("thingy")


@requires_only_gdal1
@pytest.mark.parametrize("field_type", ["time", "datetime"])
def test_check_schema_driver_support_shp(tmpdir, field_type):

    with pytest.raises(DriverSupportError):
        name = str(tmpdir.join("test_scheme.shp"))
        items = [("field1", field_type)]
        with fiona.open(
            name,
            "w",
            driver="ESRI Shapefile",
            schema={"geometry": "LineString", "properties": items},
        ) as c:
            pass


@requires_only_gdal1
def test_check_schema_driver_support_gpkg(tmpdir):
    with pytest.raises(DriverSupportError):
        name = str(tmpdir.join("test_scheme.gpkg"))
        items = [("field1", "time")]
        with fiona.open(
            name,
            "w",
            driver="GPKG",
            schema={"geometry": "LineString", "properties": items},
        ) as c:
            pass


@pytest.mark.parametrize("driver", ["GPKG", "GeoJSON"])
def test_geometry_only_schema_write(tmpdir, driver):
    schema = {
        "geometry": "Polygon",
        # No properties defined here.
    }

    record = Feature.from_dict(
        **{
            "geometry": {
                "type": "Polygon",
                "coordinates": [
                    [(0.0, 0.0), (1.0, 0.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)]
                ],
            }
        }
    )

    path = str(tmpdir.join(get_temp_filename(driver)))

    with fiona.open(path, mode="w", driver=driver, schema=schema) as c:
        c.write(record)

    with fiona.open(path, mode="r", driver=driver) as c:
        data = [f for f in c]
        assert len(data) == 1
        assert len(data[0].properties) == 0
        assert data[0].geometry.type == record.geometry["type"]


@pytest.mark.parametrize("driver", ["GPKG", "GeoJSON"])
def test_geometry_only_schema_update(tmpdir, driver):

    # Guard unsupported drivers
    if driver in driver_mode_mingdal["a"] and GDALVersion.runtime() < GDALVersion(
        *driver_mode_mingdal["a"][driver][:2]
    ):
        return

    schema = {
        "geometry": "Polygon",
        # No properties defined here.
    }

    record1 = Feature.from_dict(
        **{
            "geometry": {
                "type": "Polygon",
                "coordinates": [
                    [(0.0, 0.0), (1.0, 0.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)]
                ],
            }
        }
    )
    record2 = Feature.from_dict(
        **{
            "geometry": {
                "type": "Polygon",
                "coordinates": [
                    [(0.0, 0.0), (2.0, 0.0), (2.0, 2.0), (2.0, 0.0), (0.0, 0.0)]
                ],
            }
        }
    )

    path = str(tmpdir.join(get_temp_filename(driver)))

    # Create file
    with fiona.open(path, mode="w", driver=driver, schema=schema) as c:
        c.write(record1)

    # Append record
    with fiona.open(path, mode="a", driver=driver) as c:
        c.write(record2)

    with fiona.open(path, mode="r", driver=driver) as c:
        data = [f for f in c]
        assert len(data) == 2

        for f in data:
            assert len(f.properties) == 0

        assert data[0].geometry.type == record1.geometry["type"]
        assert data[1].geometry.type == record2.geometry["type"]


@pytest.mark.parametrize("driver", ["GPKG", "GeoJSON"])
def test_property_only_schema_write(tmpdir, driver):

    schema = {
        # No geometry defined here.
        "properties": {"prop1": "str"}
    }

    record1 = Feature.from_dict(**{"properties": {"prop1": "one"}})

    path = str(tmpdir.join(get_temp_filename(driver)))

    with fiona.open(path, mode="w", driver=driver, schema=schema) as c:
        c.write(record1)

    with fiona.open(path, mode="r", driver=driver) as c:
        data = [f for f in c]
        assert len(data) == 1
        assert len(data[0].properties) == 1
        assert "prop1" in data[0].properties and data[0].properties["prop1"] == "one"
        for f in data:
            assert f.geometry is None


@pytest.mark.parametrize("driver", ["GPKG", "GeoJSON"])
def test_property_only_schema_update(tmpdir, driver):

    # Guard unsupported drivers
    if driver in driver_mode_mingdal["a"] and GDALVersion.runtime() < GDALVersion(
        *driver_mode_mingdal["a"][driver][:2]
    ):
        return

    schema = {
        # No geometry defined here.
        "properties": {"prop1": "str"}
    }

    record1 = Feature.from_dict(**{"properties": {"prop1": "one"}})
    record2 = Feature.from_dict(**{"properties": {"prop1": "two"}})

    path = str(tmpdir.join(get_temp_filename(driver)))

    # Create file
    with fiona.open(path, mode="w", driver=driver, schema=schema) as c:
        c.write(record1)

    # Append record
    with fiona.open(path, mode="a", driver=driver) as c:
        c.write(record2)

    with fiona.open(path, mode="r", driver=driver) as c:
        data = [f for f in c]
        assert len(data) == 2
        for f in data:
            assert len(f.properties) == 1
            assert f.geometry is None
        assert "prop1" in data[0].properties and data[0].properties["prop1"] == "one"
        assert "prop1" in data[1].properties and data[1].properties["prop1"] == "two"


def test_schema_default_fields_wrong_type(tmpdir):
    """Test for SchemaError if a default field is specified with a different type"""

    name = str(tmpdir.join("test.gpx"))
    schema = {
        "properties": OrderedDict([("ele", "str"), ("time", "datetime")]),
        "geometry": "Point",
    }

    with pytest.raises(SchemaError):
        with fiona.open(name, "w", driver="GPX", schema=schema) as c:
            pass


def test_schema_string_list(tmp_path):
    output_file = tmp_path / "fio_test.geojson"
    schema = {
    "properties": {
        "time_range": "str",
    },
    "geometry": "Point",
}
    with fiona.open(
        output_file, "w", driver="GeoJSON", schema=schema, crs="EPSG:4326"
    ) as fds:
        fds.writerecords(
            [
                {
                    "id": 1,
                    "geometry": {"type": "Point", "coordinates": [0.0, 0.0]},
                    "properties": {
                        "time_range": '["2020-01-01", "2020-01-02"]',
                    },
                }
            ]
        )

    with fiona.open(output_file) as fds:
        assert fds.schema["properties"] == {"time_range": "List[str]"}
        layers = list(fds)
        assert layers[0]["properties"] == {
            "time_range": ["2020-01-01", "2020-01-02"]
        }



"""Tests for ``fiona.BytesCollection()``."""


import pytest

import fiona
from fiona.model import Geometry


class TestReading:
    @pytest.fixture(autouse=True)
    def bytes_collection_object(self, path_coutwildrnp_json):
        with open(path_coutwildrnp_json) as src:
            bytesbuf = src.read().encode("utf-8")
        self.c = fiona.BytesCollection(bytesbuf, encoding="utf-8")
        yield
        self.c.close()

    def test_construct_with_str(self, path_coutwildrnp_json):
        with open(path_coutwildrnp_json) as src:
            strbuf = src.read()
        with pytest.raises(ValueError):
            fiona.BytesCollection(strbuf)

    def test_open_repr(self):
        # I'm skipping checking the name of the virtual file as it produced by uuid.
        print(repr(self.c))
        assert repr(self.c).startswith("<open BytesCollection '/vsimem/")

    def test_closed_repr(self):
        # I'm skipping checking the name of the virtual file as it produced by uuid.
        self.c.close()
        print(repr(self.c))
        assert repr(self.c).startswith("<closed BytesCollection '/vsimem/")

    def test_path(self):
        assert self.c.path == self.c.virtual_file

    def test_closed_virtual_file(self):
        self.c.close()
        assert self.c.virtual_file is None

    def test_closed_buf(self):
        self.c.close()
        assert self.c.bytesbuf is None

    def test_name(self):
        assert len(self.c.name) > 0

    def test_mode(self):
        assert self.c.mode == "r"

    def test_collection(self):
        assert self.c.encoding == "utf-8"

    def test_iter(self):
        assert iter(self.c)

    def test_closed_no_iter(self):
        self.c.close()
        with pytest.raises(ValueError):
            iter(self.c)

    def test_len(self):
        assert len(self.c) == 67

    def test_closed_len(self):
        # Len is lazy, it's never computed in this case. TODO?
        self.c.close()
        assert len(self.c) == 0

    def test_len_closed_len(self):
        # Lazy len is computed in this case and sticks.
        len(self.c)
        self.c.close()
        assert len(self.c) == 67

    def test_driver(self):
        assert self.c.driver == "GeoJSON"

    def test_closed_driver(self):
        self.c.close()
        assert self.c.driver is None

    def test_driver_closed_driver(self):
        self.c.driver
        self.c.close()
        assert self.c.driver == "GeoJSON"

    def test_schema(self):
        s = self.c.schema["properties"]
        assert s["PERIMETER"] == "float"
        assert s["NAME"] == "str"
        assert s["URL"] == "str"
        assert s["STATE_FIPS"] == "str"
        assert s["WILDRNP020"] == "int"

    def test_closed_schema(self):
        # Schema is lazy too, never computed in this case. TODO?
        self.c.close()
        assert self.c.schema is None

    def test_schema_closed_schema(self):
        self.c.schema
        self.c.close()
        assert sorted(self.c.schema.keys()) == ["geometry", "properties"]

    def test_crs(self):
        assert self.c.crs["init"] == "epsg:4326"

    def test_crs_wkt(self):
        assert self.c.crs_wkt.startswith('GEOGCS["WGS 84"')

    def test_closed_crs(self):
        # Crs is lazy too, never computed in this case. TODO?
        self.c.close()
        assert self.c.crs is None

    def test_crs_closed_crs(self):
        self.c.crs
        self.c.close()
        assert sorted(self.c.crs.keys()) == ["init"]

    def test_meta(self):
        assert sorted(self.c.meta.keys()) == ["crs", "crs_wkt", "driver", "schema"]

    def test_bounds(self):
        assert self.c.bounds[0] == pytest.approx(-113.564247)
        assert self.c.bounds[1] == pytest.approx(37.068981)
        assert self.c.bounds[2] == pytest.approx(-104.970871)
        assert self.c.bounds[3] == pytest.approx(41.996277)

    def test_iter_one(self):
        itr = iter(self.c)
        f = next(itr)
        assert f["id"] == "0"
        assert f["properties"]["STATE"] == "UT"

    def test_iter_list(self):
        f = list(self.c)[0]
        assert f["id"] == "0"
        assert f["properties"]["STATE"] == "UT"

    def test_re_iter_list(self):
        f = list(self.c)[0]  # Run through iterator
        f = list(self.c)[0]  # Run through a new, reset iterator
        assert f["id"] == "0"
        assert f["properties"]["STATE"] == "UT"

    def test_getitem_one(self):
        f = self.c[0]
        assert f["id"] == "0"
        assert f["properties"]["STATE"] == "UT"

    def test_no_write(self):
        with pytest.raises(OSError):
            self.c.write({})

    def test_iter_items_list(self):
        i, f = list(self.c.items())[0]
        assert i == 0
        assert f["id"] == "0"
        assert f["properties"]["STATE"] == "UT"

    def test_iter_keys_list(self):
        i = list(self.c.keys())[0]
        assert i == 0

    def test_in_keys(self):
        assert 0 in self.c.keys()
        assert 0 in self.c


class TestFilterReading:
    @pytest.fixture(autouse=True)
    def bytes_collection_object(self, path_coutwildrnp_json):
        with open(path_coutwildrnp_json) as src:
            bytesbuf = src.read().encode("utf-8")
        self.c = fiona.BytesCollection(bytesbuf)
        yield
        self.c.close()

    def test_filter_1(self):
        results = list(self.c.filter(bbox=(-120.0, 30.0, -100.0, 50.0)))
        assert len(results) == 67
        f = results[0]
        assert f["id"] == "0"
        assert f["properties"]["STATE"] == "UT"

    def test_filter_reset(self):
        results = list(self.c.filter(bbox=(-112.0, 38.0, -106.0, 40.0)))
        assert len(results) == 26
        results = list(self.c.filter())
        assert len(results) == 67

    def test_filter_mask(self):
        mask = Geometry.from_dict(
            **{
                "type": "Polygon",
                "coordinates": (
                    ((-112, 38), (-112, 40), (-106, 40), (-106, 38), (-112, 38)),
                ),
            }
        )
        results = list(self.c.filter(mask=mask))
        assert len(results) == 26


def test_zipped_bytes_collection(bytes_coutwildrnp_zip):
    """Open a zipped stream of bytes as a collection"""
    with fiona.BytesCollection(bytes_coutwildrnp_zip) as col:
        assert col.name == "coutwildrnp"
        assert len(col) == 67


@pytest.mark.skipif(
    fiona.gdal_version >= (2, 3, 0),
    reason="Changed behavior with gdal 2.3, possibly related to RFC 70:"
    "Guessing output format from output file name extension for utilities",
)
def test_grenada_bytes_geojson(bytes_grenada_geojson):
    """Read grenada.geojson as BytesCollection.

    grenada.geojson is an example of geojson that GDAL's GeoJSON
    driver will fail to read successfully unless the file's extension
    reflects its json'ness.
    """
    # We expect an exception if the GeoJSON driver isn't specified.
    with pytest.raises(fiona.errors.FionaValueError):
        with fiona.BytesCollection(bytes_grenada_geojson) as col:
            pass

    # If told what driver to use, we should be good.
    with fiona.BytesCollection(bytes_grenada_geojson, driver="GeoJSON") as col:
        assert len(col) == 1



"""New tests of writing feature collections."""

import pytest

from .conftest import requires_gdal33

import fiona
from fiona.crs import CRS
from fiona.errors import DriverError
from fiona.model import Feature


def test_issue771(tmpdir, caplog):
    """Overwrite a GeoJSON file without logging errors."""
    schema = {"geometry": "Point", "properties": {"zero": "int"}}

    feature = Feature.from_dict(
        **{
            "geometry": {"type": "Point", "coordinates": (0, 0)},
            "properties": {"zero": "0"},
        }
    )

    outputfile = tmpdir.join("test.geojson")

    for i in range(2):
        with fiona.open(
            str(outputfile),
            "w",
            driver="GeoJSON",
            schema=schema,
            crs=CRS.from_epsg(4326),
        ) as collection:
            collection.write(feature)
        assert outputfile.exists()

    for record in caplog.records:
        assert record.levelname != "ERROR"


@requires_gdal33
def test_write__esri_only_wkt(tmpdir):
    """https://github.com/Toblerity/Fiona/issues/977"""
    schema = {"geometry": "Point", "properties": {"zero": "int"}}
    feature = Feature.from_dict(
        **{
            "geometry": {"type": "Point", "coordinates": (0, 0)},
            "properties": {"zero": "0"},
        }
    )
    target_crs = (
        'PROJCS["IaRCS_04_Sioux_City-Iowa_Falls_NAD_1983_2011_LCC_US_Feet",'
        'GEOGCS["GCS_NAD_1983_2011",DATUM["D_NAD_1983_2011",'
        'SPHEROID["GRS_1980",6378137.0,298.257222101]],'
        'PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],'
        'PROJECTION["Lambert_Conformal_Conic"],'
        'PARAMETER["False_Easting",14500000.0],'
        'PARAMETER["False_Northing",8600000.0],'
        'PARAMETER["Central_Meridian",-94.83333333333333],'
        'PARAMETER["Standard_Parallel_1",42.53333333333333],'
        'PARAMETER["Standard_Parallel_2",42.53333333333333],'
        'PARAMETER["Scale_Factor",1.000045],'
        'PARAMETER["Latitude_Of_Origin",42.53333333333333],'
        'UNIT["Foot_US",0.3048006096012192]]'
    )
    outputfile = tmpdir.join("test.shp")
    with fiona.open(
        str(outputfile),
        "w",
        driver="ESRI Shapefile",
        schema=schema,
        crs=target_crs,
    ) as collection:
        collection.write(feature)
        assert collection.crs_wkt.startswith(
            (
                'PROJCS["IaRCS_04_Sioux_City-Iowa_Falls_NAD_1983_2011_LCC_US_Feet"',
                'PROJCRS["IaRCS_04_Sioux_City-Iowa_Falls_NAD_1983_2011_LCC_US_Feet"',  # GDAL 3.3+
            )
        )


def test_write__wkt_version(tmpdir):
    """https://github.com/Toblerity/Fiona/issues/977"""
    schema = {"geometry": "Point", "properties": {"zero": "int"}}
    feature = Feature.from_dict(
        **{
            "geometry": {"type": "Point", "coordinates": (0, 0)},
            "properties": {"zero": "0"},
        }
    )
    target_crs = (
        'PROJCS["IaRCS_04_Sioux_City-Iowa_Falls_NAD_1983_2011_LCC_US_Feet",'
        'GEOGCS["GCS_NAD_1983_2011",DATUM["D_NAD_1983_2011",'
        'SPHEROID["GRS_1980",6378137.0,298.257222101]],'
        'PRIMEM["Greenwich",0.0],UNIT["Degree",0.0174532925199433]],'
        'PROJECTION["Lambert_Conformal_Conic"],'
        'PARAMETER["False_Easting",14500000.0],'
        'PARAMETER["False_Northing",8600000.0],'
        'PARAMETER["Central_Meridian",-94.83333333333333],'
        'PARAMETER["Standard_Parallel_1",42.53333333333333],'
        'PARAMETER["Standard_Parallel_2",42.53333333333333],'
        'PARAMETER["Scale_Factor",1.000045],'
        'PARAMETER["Latitude_Of_Origin",42.53333333333333],'
        'UNIT["Foot_US",0.3048006096012192]]'
    )
    outputfile = tmpdir.join("test.shp")
    with fiona.open(
        str(outputfile),
        "w",
        driver="ESRI Shapefile",
        schema=schema,
        crs=target_crs,
        wkt_version="WKT2_2018",
    ) as collection:
        collection.write(feature)
        assert collection.crs_wkt.startswith(
            'PROJCRS["IaRCS_04_Sioux_City-Iowa_Falls_NAD_1983_2011_LCC_US_Feet"'
        )


def test_issue1169():
    """Don't swallow errors when a collection can't be written."""
    with pytest.raises(DriverError):
        with fiona.open(
            "s3://non-existing-bucket/test.geojson",
            mode="w",
            driver="GeoJSON",
            schema={"geometry": "Point"},
        ) as collection:
            collection.writerecords(
                [
                    Feature.from_dict(
                        **{
                            "id": "0",
                            "type": "Feature",
                            "geometry": {"type": "Point", "coordinates": (1.0, 2.0)},
                        }
                    )
                ]
            )



"""Fiona's command line interface"""

from functools import wraps


def with_context_env(f):
    """Pops the Fiona Env from the passed context and executes the
    wrapped func in the context of that obj.

    Click's pass_context decorator must precede this decorator, or else
    there will be no context in the wrapper args.
    """
    @wraps(f)
    def wrapper(*args, **kwds):
        ctx = args[0]
        env = ctx.obj.pop('env')
        with env:
            return f(*args, **kwds)
    return wrapper



"""Tests for `$ fio bounds`."""


import re

from fiona.fio import bounds
from fiona.fio.main import main_group


def test_fail(runner):
    result = runner.invoke(main_group, ['bounds', ], '5')
    assert result.exit_code == 1


def test_seq(feature_seq, runner):
    result = runner.invoke(main_group, ['bounds', ], feature_seq)
    assert result.exit_code == 0
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d*', result.output)) == 8


def test_seq_rs(feature_seq_pp_rs, runner):
    result = runner.invoke(main_group, ['bounds', ], feature_seq_pp_rs)
    assert result.exit_code == 0
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d*', result.output)) == 8


def test_precision(feature_seq, runner):
    result = runner.invoke(main_group, ['bounds', '--precision', 1], feature_seq)
    assert result.exit_code == 0
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d{1}\D', result.output)) == 8


def test_explode(feature_collection, runner):
    result = runner.invoke(main_group, ['bounds', '--explode'], feature_collection)
    assert result.exit_code == 0
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d*', result.output)) == 8


def test_explode_pp(feature_collection_pp, runner):
    result = runner.invoke(main_group, ['bounds', '--explode'], feature_collection_pp)
    assert result.exit_code == 0
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d*', result.output)) == 8


def test_with_id(feature_seq, runner):
    result = runner.invoke(main_group, ['bounds', '--with-id'], feature_seq)
    assert result.exit_code == 0
    assert result.output.count('id') == result.output.count('bbox') == 2


def test_explode_with_id(feature_collection, runner):
    result = runner.invoke(
        main_group, ['bounds', '--explode', '--with-id'], feature_collection)
    assert result.exit_code == 0
    assert result.output.count('id') == result.output.count('bbox') == 2


def test_with_obj(feature_seq, runner):
    result = runner.invoke(main_group, ['bounds', '--with-obj'], feature_seq)
    assert result.exit_code == 0
    assert result.output.count('geometry') == result.output.count('bbox') == 2


def test_bounds_explode_with_obj(feature_collection, runner):
    result = runner.invoke(
        main_group, ['bounds', '--explode', '--with-obj'], feature_collection)
    assert result.exit_code == 0
    assert result.output.count('geometry') == result.output.count('bbox') == 2


def test_explode_output_rs(feature_collection, runner):
    result = runner.invoke(main_group, ['bounds', '--explode', '--rs'], feature_collection)
    assert result.exit_code == 0
    assert result.output.count('\x1e') == 2
    assert result.output.count('[') == result.output.count(']') == 2
    assert len(re.findall(r'\d*\.\d*', result.output)) == 8



#  tests for fiona-1.9.1-py38h07fba90_0 (this is a generated file);
print('===== testing package: fiona-1.9.1-py38h07fba90_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import os
import fiona
import fiona.fio

fname = os.path.join(os.environ['RECIPE_DIR'], 'test_data', 'test.shp')

with fiona.open(fname, 'r') as f:
    assert f.schema['geometry'] == 'Point'

# https://github.com/conda-forge/fiona-feedstock/issues/49
meta = {'driver': 'ESRI Shapefile', 'schema': {'geometry': 'Point', 'properties': {'RETURN_P': 'int'}}}
dst = fiona.open('new.shp', 'w', **meta)
g = {'coordinates': [1.0, 2.0], 'type': 'Point'}
feature = {'geometry': g, 'properties': {'RETURN_P': None}}
dst.write(feature)
dst.close()

src = fiona.open('new.shp', 'r')
feature = next(src)
assert(feature['properties']['RETURN_P'] is None)  # Fails
src.close()
#  --- run_test.py (end) ---

print('===== fiona-1.9.1-py38h07fba90_0 OK =====');



"""$ fio filter"""

import json

import click
from cligj import use_rs_opt

from fiona.fio.helpers import obj_gen, eval_feature_expression
from fiona.fio import with_context_env


@click.command()
@click.argument('filter_expression')
@use_rs_opt
@click.pass_context
@with_context_env
def filter(ctx, filter_expression, use_rs):
    """
    Filter GeoJSON features by python expression.

    Features are read from stdin.

    The expression is evaluated in a restricted namespace containing:
        - sum, pow, min, max and the imported math module
        - shape (optional, imported from shapely.geometry if available)
        - bool, int, str, len, float type conversions
        - f (the feature to be evaluated,
             allows item access via javascript-style dot notation using munch)

    The expression will be evaluated for each feature and, if true,
    the feature will be included in the output.  For example:

    \b
        $ fio cat data.shp \\
            | fio filter "f.properties.area > 1000.0" \\
            | fio collect > large_polygons.geojson

    """
    stdin = click.get_text_stream('stdin')
    source = obj_gen(stdin)

    for i, obj in enumerate(source):
        features = obj.get("features") or [obj]
        for j, feat in enumerate(features):
            if not eval_feature_expression(feat, filter_expression):
                continue

            if use_rs:
                click.echo("\x1e", nl=False)
            click.echo(json.dumps(feat))



"""Tests of the transform submodule"""

import math
import pytest
from fiona import transform
from fiona.errors import FionaDeprecationWarning
from fiona.model import Geometry

from .conftest import requires_gdal_lt_3


TEST_GEOMS = [
    Geometry(type="Point", coordinates=[0.0, 0.0, 1000.0]),
    Geometry(type="LineString", coordinates=[[0.0, 0.0, 1000.0], [0.1, 0.1, -1000.0]]),
    Geometry(type="MultiPoint", coordinates=[[0.0, 0.0, 1000.0], [0.1, 0.1, -1000.0]]),
    Geometry(
        type="Polygon",
        coordinates=[
            [
                [0.0, 0.0, 1000.0],
                [0.1, 0.1, -1000.0],
                [0.1, -0.1, math.pi],
                [0.0, 0.0, 1000.0],
            ]
        ],
    ),
    Geometry(
        type="MultiPolygon",
        coordinates=[
            [
                [
                    [0.0, 0.0, 1000.0],
                    [0.1, 0.1, -1000.0],
                    [0.1, -0.1, math.pi],
                    [0.0, 0.0, 1000.0],
                ]
            ]
        ],
    ),
]


@pytest.mark.parametrize("geom", TEST_GEOMS)
def test_transform_geom_with_z(geom):
    """Transforming a geom with Z succeeds"""
    transform.transform_geom("epsg:4326", "epsg:3857", geom)


@pytest.mark.parametrize("geom", TEST_GEOMS)
def test_transform_geom_array_z(geom):
    """Transforming a geom array with Z succeeds"""
    g2 = transform.transform_geom(
        "epsg:4326",
        "epsg:3857",
        [geom for _ in range(5)],
    )
    assert isinstance(g2, list)
    assert len(g2) == 5


@pytest.mark.parametrize(
    "crs",
    [
        "epsg:4326",
        "EPSG:4326",
        "WGS84",
        {"init": "epsg:4326"},
        {"proj": "longlat", "datum": "WGS84", "no_defs": True},
        "OGC:CRS84",
    ],
)
def test_axis_ordering_rev(crs):
    """Test if transform uses traditional_axis_mapping"""
    expected = (-8427998.647958742, 4587905.27136252)
    t1 = transform.transform(crs, "epsg:3857", [-75.71], [38.06])
    assert (t1[0][0], t1[1][0]) == pytest.approx(expected)
    geom = Geometry.from_dict(**{"type": "Point", "coordinates": [-75.71, 38.06]})
    g1 = transform.transform_geom(crs, "epsg:3857", geom)
    assert g1["coordinates"] == pytest.approx(expected)


@pytest.mark.parametrize(
    "crs",
    [
        "epsg:4326",
        "EPSG:4326",
        "WGS84",
        {"init": "epsg:4326"},
        {"proj": "longlat", "datum": "WGS84", "no_defs": True},
        "OGC:CRS84",
    ],
)
def test_axis_ordering_fwd(crs):
    """Test if transform uses traditional_axis_mapping"""
    rev_expected = (-75.71, 38.06)
    t2 = transform.transform("epsg:3857", crs, [-8427998.647958742], [4587905.27136252])
    assert (t2[0][0], t2[1][0]) == pytest.approx(rev_expected)
    geom = Geometry.from_dict(
        **{"type": "Point", "coordinates": [-8427998.647958742, 4587905.27136252]}
    )
    g2 = transform.transform_geom("epsg:3857", crs, geom)
    assert g2.coordinates == pytest.approx(rev_expected)


def test_transform_issue971():
    """See https://github.com/Toblerity/Fiona/issues/971"""
    source_crs = "EPSG:25832"
    dest_src = "EPSG:4326"
    geom = {
        "type": "GeometryCollection",
        "geometries": [
            {
                "type": "LineString",
                "coordinates": [
                    (512381.8870945257, 5866313.311218272),
                    (512371.23869999964, 5866322.282500001),
                    (512364.6014999999, 5866328.260199999),
                ],
            }
        ],
    }
    geom_transformed = transform.transform_geom(source_crs, dest_src, geom)
    assert geom_transformed.geometries[0].coordinates[0] == pytest.approx(
        (9.18427, 52.94630)
    )


def test_transform_geom_precision_deprecation():
    """Get a precision deprecation warning in 1.9."""
    with pytest.warns(FionaDeprecationWarning):
        transform.transform_geom(
            "epsg:4326",
            "epsg:3857",
            Geometry(type="Point", coordinates=(0, 0)),
            precision=2,
        )



"""Tests for geometry objects."""

import pytest

from fiona._geometry import GeomBuilder, geometryRT
from fiona.errors import UnsupportedGeometryTypeError
from fiona.model import Geometry


def geometry_wkb(wkb):
    try:
        wkb = bytes.fromhex(wkb)
    except AttributeError:
        wkb = wkb.decode("hex")
    return GeomBuilder().build_wkb(wkb)


def test_ogr_builder_exceptions():
    geom = Geometry.from_dict(**{"type": "Bogus", "coordinates": None})
    with pytest.raises(UnsupportedGeometryTypeError):
        geometryRT(geom)


@pytest.mark.parametrize(
    "geom_type, coordinates",
    [
        ("Point", (0.0, 0.0)),
        ("LineString", [(0.0, 0.0), (1.0, 1.0)]),
        ("Polygon", [[(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)]]),
        ("MultiPoint", [(0.0, 0.0), (1.0, 1.0)]),
        ("MultiLineString", [[(0.0, 0.0), (1.0, 1.0)]]),
        (
            "MultiPolygon",
            [[[(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0), (0.0, 0.0)]]],
        ),
    ],
)
def test_round_tripping(geom_type, coordinates):
    result = geometryRT(
        Geometry.from_dict(**{"type": geom_type, "coordinates": coordinates})
    )
    assert result.type == geom_type
    assert result.coordinates == coordinates


@pytest.mark.parametrize(
    "geom_type, coordinates",
    [
        ("Polygon", [[(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)]]),
        ("MultiPolygon", [[[(0.0, 0.0), (0.0, 1.0), (1.0, 1.0), (1.0, 0.0)]]]),
    ],
)
def test_implicitly_closed_round_tripping(geom_type, coordinates):
    result = geometryRT(
        Geometry.from_dict(**{"type": geom_type, "coordinates": coordinates})
    )
    assert result.type == geom_type
    result_coordinates = result.coordinates
    while not isinstance(coordinates[0], tuple):
        result_coordinates = result_coordinates[0]
        coordinates = coordinates[0]
    assert result_coordinates[:-1] == coordinates


def test_geometry_collection_round_trip():
    geom = {
        "type": "GeometryCollection",
        "geometries": [
            {"type": "Point", "coordinates": (0.0, 0.0)},
            {"type": "LineString", "coordinates": [(0.0, 0.0), (1.0, 1.0)]},
        ],
    }

    result = geometryRT(geom)
    assert len(result["geometries"]) == 2
    assert [g["type"] for g in result["geometries"]] == ["Point", "LineString"]


def test_point_wkb():
    # Hex-encoded Point (0 0)
    wkb = "010100000000000000000000000000000000000000"
    geom = geometry_wkb(wkb)
    assert geom["type"] == "Point"
    assert geom["coordinates"] == (0.0, 0.0)


def test_line_wkb():
    # Hex-encoded LineString (0 0, 1 1)
    wkb = (
        "01020000000200000000000000000000000000000000000000000000000000f03f"
        "000000000000f03f"
    )
    geom = geometry_wkb(wkb)
    assert geom["type"] == "LineString"
    assert geom["coordinates"] == [(0.0, 0.0), (1.0, 1.0)]


def test_polygon_wkb():
    # 1 x 1 box (0, 0, 1, 1)
    wkb = (
        "01030000000100000005000000000000000000f03f000000000000000000000000"
        "0000f03f000000000000f03f0000000000000000000000000000f03f0000000000"
        "0000000000000000000000000000000000f03f0000000000000000"
    )
    geom = geometry_wkb(wkb)
    assert geom["type"], "Polygon"
    assert len(geom["coordinates"]) == 1
    assert len(geom["coordinates"][0]) == 5
    x, y = zip(*geom["coordinates"][0])
    assert min(x) == 0.0
    assert min(y) == 0.0
    assert max(x) == 1.0
    assert max(y) == 1.0


def test_multipoint_wkb():
    wkb = (
        "010400000002000000010100000000000000000000000000000000000000010100"
        "0000000000000000f03f000000000000f03f"
    )
    geom = geometry_wkb(wkb)
    assert geom["type"] == "MultiPoint"
    assert geom["coordinates"] == [(0.0, 0.0), (1.0, 1.0)]


def test_multilinestring_wkb():
    # Hex-encoded LineString (0 0, 1 1)
    wkb = (
        "010500000001000000010200000002000000000000000000000000000000000000"
        "00000000000000f03f000000000000f03f"
    )
    geom = geometry_wkb(wkb)
    assert geom["type"] == "MultiLineString"
    assert len(geom["coordinates"]) == 1
    assert len(geom["coordinates"][0]) == 2
    assert geom["coordinates"][0] == [(0.0, 0.0), (1.0, 1.0)]


def test_multipolygon_wkb():
    # [1 x 1 box (0, 0, 1, 1)]
    wkb = (
        "01060000000100000001030000000100000005000000000000000000f03f000000"
        "0000000000000000000000f03f000000000000f03f000000000000000000000000"
        "0000f03f00000000000000000000000000000000000000000000f03f0000000000"
        "000000"
    )
    geom = geometry_wkb(wkb)
    assert geom["type"] == "MultiPolygon"
    assert len(geom["coordinates"]) == 1
    assert len(geom["coordinates"][0]) == 1
    assert len(geom["coordinates"][0][0]) == 5
    x, y = zip(*geom["coordinates"][0][0])
    assert min(x) == 0.0
    assert min(y) == 0.0
    assert max(x) == 1.0
    assert max(y) == 1.0



"""Tests for `$ fio distrib`."""


from click.testing import CliRunner

from fiona.fio.main import main_group


def test_distrib(feature_collection_pp):
    runner = CliRunner()
    result = runner.invoke(main_group, ['distrib', ], feature_collection_pp)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 2


def test_distrib_no_rs(feature_collection):
    runner = CliRunner()
    result = runner.invoke(main_group, ['distrib', ], feature_collection)
    assert result.exit_code == 0
    assert result.output.count('"Feature"') == 2



"""Tests of GDAL and PROJ data finding"""

import os.path

from click.testing import CliRunner
import pytest

import fiona
from fiona._env import GDALDataFinder, PROJDataFinder
from fiona.fio.main import main_group


@pytest.mark.wheel
def test_gdal_data_wheel():
    """Get GDAL data path from a wheel"""
    assert GDALDataFinder().search() == os.path.join(os.path.dirname(fiona.__file__), 'gdal_data')


@pytest.mark.wheel
def test_proj_data_wheel():
    """Get GDAL data path from a wheel"""
    assert PROJDataFinder().search() == os.path.join(os.path.dirname(fiona.__file__), 'proj_data')


@pytest.mark.wheel
def test_env_gdal_data_wheel():
    runner = CliRunner()
    result = runner.invoke(main_group, ['env', '--gdal-data'])
    assert result.exit_code == 0
    assert result.output.strip() == os.path.join(os.path.dirname(fiona.__file__), 'gdal_data')


@pytest.mark.wheel
def test_env_proj_data_wheel():
    runner = CliRunner()
    result = runner.invoke(main_group, ['env', '--proj-data'])
    assert result.exit_code == 0
    assert result.output.strip() == os.path.join(os.path.dirname(fiona.__file__), 'proj_data')


def test_env_gdal_data_environ(monkeypatch):
    monkeypatch.setenv('GDAL_DATA', '/foo/bar')
    runner = CliRunner()
    result = runner.invoke(main_group, ['env', '--gdal-data'])
    assert result.exit_code == 0
    assert result.output.strip() == '/foo/bar'


@pytest.mark.parametrize("data_directory_env", ["PROJ_LIB", "PROJ_DATA"])
def test_env_proj_data_environ(data_directory_env, monkeypatch):
    monkeypatch.delenv('PROJ_DATA', raising=False)
    monkeypatch.delenv('PROJ_LIB', raising=False)
    monkeypatch.setenv(data_directory_env, '/foo/bar')
    runner = CliRunner()
    result = runner.invoke(main_group, ['env', '--proj-data'])
    assert result.exit_code == 0
    assert result.output.strip() == '/foo/bar'



"""fio-cat"""

import json
import warnings

import click
import cligj

import fiona
from fiona.transform import transform_geom
from fiona.model import Feature, ObjectEncoder
from fiona.fio import options, with_context_env
from fiona.fio.helpers import recursive_round
from fiona.errors import AttributeFilterError

warnings.simplefilter("default")


# Cat command
@click.command(short_help="Concatenate and print the features of datasets")
@click.argument("files", nargs=-1, required=True, metavar="INPUTS...")
@click.option(
    "--layer",
    default=None,
    multiple=True,
    callback=options.cb_multilayer,
    help="Input layer(s), specified as 'fileindex:layer` "
    "For example, '1:foo,2:bar' will concatenate layer foo "
    "from file 1 and layer bar from file 2",
)
@cligj.precision_opt
@cligj.indent_opt
@cligj.compact_opt
@click.option(
    "--ignore-errors/--no-ignore-errors",
    default=False,
    help="log errors but do not stop serialization.",
)
@options.dst_crs_opt
@cligj.use_rs_opt
@click.option(
    "--bbox",
    default=None,
    metavar="w,s,e,n",
    help="filter for features intersecting a bounding box",
)
@click.option(
    "--where",
    default=None,
    help="attribute filter using SQL where clause",
)
@click.option(
    "--cut-at-antimeridian",
    is_flag=True,
    default=False,
    help="Optionally cut geometries at the anti-meridian. To be used only for a geographic destination CRS.",
)
@click.pass_context
@with_context_env
def cat(
    ctx,
    files,
    precision,
    indent,
    compact,
    ignore_errors,
    dst_crs,
    use_rs,
    bbox,
    where,
    cut_at_antimeridian,
    layer,
):
    """
    Concatenate and print the features of input datasets as a sequence of
    GeoJSON features.

    When working with a multi-layer dataset the first layer is used by default.
    Use the '--layer' option to select a different layer.

    """
    dump_kwds = {"sort_keys": True}
    if indent:
        dump_kwds["indent"] = indent
    if compact:
        dump_kwds["separators"] = (",", ":")

    # Validate file idexes provided in --layer option
    # (can't pass the files to option callback)
    if layer:
        options.validate_multilayer_file_index(files, layer)

    # first layer is the default
    for i in range(1, len(files) + 1):
        if str(i) not in layer.keys():
            layer[str(i)] = [0]

    try:
        if bbox:
            try:
                bbox = tuple(map(float, bbox.split(",")))
            except ValueError:
                bbox = json.loads(bbox)

        for i, path in enumerate(files, 1):
            for lyr in layer[str(i)]:
                with fiona.open(path, layer=lyr) as src:
                    for i, feat in src.items(bbox=bbox, where=where):
                        geom = feat.geometry

                        if dst_crs:
                            geom = transform_geom(
                                src.crs,
                                dst_crs,
                                geom,
                                antimeridian_cutting=cut_at_antimeridian,
                            )

                        if precision >= 0:
                            geom = recursive_round(geom, precision)

                        feat = Feature(
                            id=feat.id,
                            properties=feat.properties,
                            geometry=geom,
                            bbox=fiona.bounds(geom),
                        )

                        if use_rs:
                            click.echo("\x1e", nl=False)

                        click.echo(json.dumps(feat, cls=ObjectEncoder, **dump_kwds))

    except AttributeFilterError as e:
        raise click.BadParameter("'where' clause is invalid: " + str(e))



from collections import OrderedDict

import pytest

import fiona
from fiona.model import Feature, Geometry, Properties


class TestReadAccess:
    # To check that we'll be able to get multiple 'r' connections to layers
    # in a single file.

    def test_meta(self, path_coutwildrnp_shp):
        with fiona.open(path_coutwildrnp_shp, "r", layer="coutwildrnp") as c:
            with fiona.open(path_coutwildrnp_shp, "r", layer="coutwildrnp") as c2:
                assert len(c) == len(c2)
                assert sorted(c.schema.items()) == sorted(c2.schema.items())

    def test_feat(self, path_coutwildrnp_shp):
        with fiona.open(path_coutwildrnp_shp, "r", layer="coutwildrnp") as c:
            f1 = next(iter(c))
            with fiona.open(path_coutwildrnp_shp, "r", layer="coutwildrnp") as c2:
                f2 = next(iter(c2))
                assert f1.id == f2.id
                assert f1.properties == f2.properties
                assert f1.geometry.type == f2.geometry.type


class TestReadWriteAccess:
    # To check that we'll be able to read from a file that we're
    # writing to.

    @pytest.fixture(autouse=True)
    def multi_write_test_shp(self, tmpdir):
        self.shapefile_path = str(tmpdir.join("multi_write_test.shp"))
        self.c = fiona.open(
            self.shapefile_path,
            "w",
            driver="ESRI Shapefile",
            schema={
                "geometry": "Point",
                "properties": [("title", "str:80"), ("date", "date")],
            },
            crs={"init": "epsg:4326", "no_defs": True},
            encoding="utf-8",
        )
        self.f = Feature(
            id="0",
            geometry=Geometry(type="Point", coordinates=(0.0, 0.1)),
            properties=Properties(title="point one", date="2012-01-29"),
        )
        self.c.writerecords([self.f])
        self.c.flush()
        yield
        self.c.close()

    def test_meta(self):
        c2 = fiona.open(self.shapefile_path, "r")
        assert len(self.c) == len(c2)
        assert sorted(self.c.schema.items()) == sorted(c2.schema.items())
        c2.close()

    def test_read(self):
        c2 = fiona.open(self.shapefile_path, "r")
        f2 = next(iter(c2))
        assert self.f.id == f2.id
        assert self.f.properties == f2.properties
        assert self.f.geometry.type == f2.geometry.type
        c2.close()

    def test_read_after_close(self):
        c2 = fiona.open(self.shapefile_path, "r")
        self.c.close()
        f2 = next(iter(c2))
        assert self.f.properties == f2.properties
        c2.close()


class TestLayerCreation:
    @pytest.fixture(autouse=True)
    def layer_creation_shp(self, tmpdir):
        self.dir = tmpdir.mkdir("layer_creation")
        self.c = fiona.open(
            str(self.dir),
            "w",
            layer="write_test",
            driver="ESRI Shapefile",
            schema={
                "geometry": "Point",
                "properties": [("title", "str:80"), ("date", "date")],
            },
            crs={"init": "epsg:4326", "no_defs": True},
            encoding="utf-8",
        )
        self.f = Feature(
            geometry=Geometry(type="Point", coordinates=(0.0, 0.1)),
            properties=OrderedDict([("title", "point one"), ("date", "2012-01-29")]),
        )
        self.c.writerecords([self.f])
        self.c.flush()
        yield
        self.c.close()

    def test_meta(self):
        c2 = fiona.open(str(self.dir.join("write_test.shp")), "r")
        assert len(self.c) == len(c2)
        assert sorted(self.c.schema.items()) == sorted(c2.schema.items())
        c2.close()

    def test_read(self):
        c2 = fiona.open(str(self.dir.join("write_test.shp")), "r")
        f2 = next(iter(c2))
        assert self.f.properties == f2.properties
        c2.close()

    def test_read_after_close(self):
        c2 = fiona.open(str(self.dir.join("write_test.shp")), "r")
        self.c.close()
        f2 = next(iter(c2))
        assert self.f.properties == f2.properties
        c2.close()



"""Test of deprecations following RFC 1"""

import pytest

from fiona.errors import FionaDeprecationWarning
from fiona.model import (
    _Geometry,
    Feature,
    Geometry,
    Object,
    ObjectEncoder,
    Properties,
    decode_object,
)


def test_object_len():
    """object len is correct"""
    obj = Object(g=1)
    assert len(obj) == 1


def test_object_iter():
    """object iter is correct"""
    obj = Object(g=1)
    assert [obj[k] for k in obj] == [1]


def test_object_setitem_warning():
    """Warn about __setitem__"""
    obj = Object()
    with pytest.warns(FionaDeprecationWarning, match="immutable"):
        obj["g"] = 1
    assert "g" in obj
    assert obj["g"] == 1


def test_object_update_warning():
    """Warn about update"""
    obj = Object()
    with pytest.warns(FionaDeprecationWarning, match="immutable"):
        obj.update(g=1)
    assert "g" in obj
    assert obj["g"] == 1


def test_object_popitem_warning():
    """Warn about pop"""
    obj = Object(g=1)
    with pytest.warns(FionaDeprecationWarning, match="immutable"):
        assert obj.pop("g") == 1
    assert "g" not in obj


def test_object_delitem_warning():
    """Warn about __delitem__"""
    obj = Object(g=1)
    with pytest.warns(FionaDeprecationWarning, match="immutable"):
        del obj["g"]
    assert "g" not in obj


def test_object_setitem_delegated():
    """Delegation in __setitem__ works"""

    class ThingDelegate:
        def __init__(self, value):
            self.value = value

    class Thing(Object):
        _delegated_properties = ["value"]

        def __init__(self, value=None, **data):
            self._delegate = ThingDelegate(value)
            super().__init__(**data)

    thing = Thing()
    assert thing["value"] is None
    with pytest.warns(FionaDeprecationWarning, match="immutable"):
        thing["value"] = 1
    assert thing["value"] == 1


def test_object_delitem_delegated():
    """Delegation in __delitem__ works"""

    class ThingDelegate:
        def __init__(self, value):
            self.value = value

    class Thing(Object):
        _delegated_properties = ["value"]

        def __init__(self, value=None, **data):
            self._delegate = ThingDelegate(value)
            super().__init__(**data)

    thing = Thing(1)
    assert thing["value"] == 1
    with pytest.warns(FionaDeprecationWarning, match="immutable"):
        del thing["value"]
    assert thing["value"] is None


def test__geometry_ctor():
    """Construction of a _Geometry works"""
    geom = _Geometry(type="Point", coordinates=(0, 0))
    assert geom.type == "Point"
    assert geom.coordinates == (0, 0)


def test_geometry_type():
    """Geometry has a type"""
    geom = Geometry(type="Point")
    assert geom.type == "Point"


def test_geometry_coordinates():
    """Geometry has coordinates"""
    geom = Geometry(coordinates=[(0, 0), (1, 1)])
    assert geom.coordinates == [(0, 0), (1, 1)]


def test_geometry__props():
    """Geometry properties as a dict"""
    assert Geometry(coordinates=(0, 0), type="Point")._props() == {
        "coordinates": (0, 0),
        "type": "Point",
        "geometries": None,
    }


def test_geometry_gi():
    """Geometry __geo_interface__"""
    gi = Geometry(coordinates=(0, 0), type="Point").__geo_interface__
    assert gi["type"] == "Point"
    assert gi["coordinates"] == (0, 0)


def test_feature_no_geometry():
    """Feature has no attribute"""
    feat = Feature()
    assert feat.geometry is None


def test_feature_geometry():
    """Feature has a geometry attribute"""
    geom = Geometry(type="Point")
    feat = Feature(geometry=geom)
    assert feat.geometry is geom


def test_feature_no_id():
    """Feature has no id"""
    feat = Feature()
    assert feat.id is None


def test_feature_id():
    """Feature has an id"""
    feat = Feature(id="123")
    assert feat.id == "123"


def test_feature_no_properties():
    """Feature has no properties"""
    feat = Feature()
    assert len(feat.properties) == 0


def test_feature_properties():
    """Feature has properties"""
    feat = Feature(properties=Properties(foo=1))
    assert len(feat.properties) == 1
    assert feat.properties["foo"] == 1


def test_feature_from_dict_kwargs():
    """Feature can be created from GeoJSON kwargs"""
    data = {
        "id": "foo",
        "type": "Feature",
        "geometry": {"type": "Point", "coordinates": (0, 0)},
        "properties": {"a": 0, "b": "bar"},
        "extras": {"this": 1},
    }
    feat = Feature.from_dict(**data)
    assert feat.id == "foo"
    assert feat.type == "Feature"
    assert feat.geometry.type == "Point"
    assert feat.geometry.coordinates == (0, 0)
    assert len(feat.properties) == 2
    assert feat.properties["a"] == 0
    assert feat.properties["b"] == "bar"
    assert feat["extras"]["this"] == 1


def test_feature_from_dict_obj():
    """Feature can be created from GeoJSON obj"""
    data = {
        "id": "foo",
        "type": "Feature",
        "geometry": {"type": "Point", "coordinates": (0, 0)},
        "properties": {"a": 0, "b": "bar"},
        "extras": {"this": 1},
    }
    feat = Feature.from_dict(data)
    assert feat.id == "foo"
    assert feat.type == "Feature"
    assert feat.geometry.type == "Point"
    assert feat.geometry.coordinates == (0, 0)
    assert len(feat.properties) == 2
    assert feat.properties["a"] == 0
    assert feat.properties["b"] == "bar"
    assert feat["extras"]["this"] == 1


def test_feature_from_dict_kwargs_2():
    """From GeoJSON kwargs using Geometry and Properties"""
    data = {
        "id": "foo",
        "type": "Feature",
        "geometry": Geometry(type="Point", coordinates=(0, 0)),
        "properties": Properties(a=0, b="bar"),
        "extras": {"this": 1},
    }
    feat = Feature.from_dict(**data)
    assert feat.id == "foo"
    assert feat.type == "Feature"
    assert feat.geometry.type == "Point"
    assert feat.geometry.coordinates == (0, 0)
    assert len(feat.properties) == 2
    assert feat.properties["a"] == 0
    assert feat.properties["b"] == "bar"
    assert feat["extras"]["this"] == 1


def test_geometry_encode():
    """Can encode a geometry"""
    assert ObjectEncoder().default(Geometry(type="Point", coordinates=(0, 0))) == {
        "type": "Point",
        "coordinates": (0, 0),
    }


@pytest.mark.parametrize("value", [100, "foo"])
def test_encode_error(value):
    """Raises TypeError"""
    with pytest.raises(TypeError):
        ObjectEncoder().default(value)


def test_feature_encode():
    """Can encode a feature"""
    o_dict = ObjectEncoder().default(
        Feature(
            id="foo",
            geometry=Geometry(type="Point", coordinates=(0, 0)),
            properties=Properties(a=1, foo="bar"),
        )
    )
    assert o_dict["id"] == "foo"
    assert o_dict["geometry"]["type"] == "Point"
    assert o_dict["geometry"]["coordinates"] == (0, 0)


def test_decode_object_hook():
    """Can decode a feature"""
    data = {
        "id": "foo",
        "type": "Feature",
        "geometry": {"type": "Point", "coordinates": (0, 0)},
        "properties": {"a": 0, "b": "bar"},
        "extras": {"this": 1},
    }
    feat = decode_object(data)
    assert feat.id == "foo"
    assert feat.type == "Feature"
    assert feat.geometry.type == "Point"
    assert feat.geometry.coordinates == (0, 0)
    assert len(feat.properties) == 2
    assert feat.properties["a"] == 0
    assert feat.properties["b"] == "bar"
    assert feat["extras"]["this"] == 1


def test_decode_object_hook_geometry():
    """Can decode a geometry"""
    data = {"type": "Point", "coordinates": (0, 0)}
    geometry = decode_object(data)
    assert geometry.type == "Point"
    assert geometry.coordinates == (0, 0)


@pytest.mark.parametrize("o", [{}, {"a": 1}, {"type": "FeatureCollection"}])
def test_decode_object_hook_fallback(o):
    """Pass through an ordinary dict"""
    assert decode_object(o) == o


def test_properties():
    """Property factory works"""
    assert Properties.from_dict(a=1, foo="bar")["a"] == 1


def test_feature_gi():
    """Feature __geo_interface__."""
    gi = Feature(
        id="foo",
        geometry=Geometry(type="Point", coordinates=(0, 0)),
        properties=Properties(a=1, foo="bar"),
    )

    assert gi["id"] == "foo"
    assert gi["geometry"]["type"] == "Point"
    assert gi["geometry"]["coordinates"] == (0, 0)




cdef extern from "cpl_error.h":

    ctypedef enum CPLErr:
        CE_None
        CE_Debug
        CE_Warning
        CE_Failure
        CE_Fatal

    int CPLGetLastErrorNo()
    const char* CPLGetLastErrorMsg()
    int CPLGetLastErrorType()
    void CPLErrorReset()


cdef extern from "ogr_srs_api.h":

    void OSRSetPROJSearchPaths(const char *const *papszPaths)
    void OSRGetPROJVersion	(int *pnMajor, int *pnMinor, int *pnPatch)



"""Helper objects needed by multiple CLI commands.

"""

from functools import partial
import json
import math
import warnings

from munch import munchify

from fiona.model import Geometry, to_dict


warnings.simplefilter("default")


def obj_gen(lines, object_hook=None):
    """Return a generator of JSON objects loaded from ``lines``."""
    first_line = next(lines)
    if first_line.startswith("\x1e"):

        def gen():
            buffer = first_line.strip("\x1e")
            for line in lines:
                if line.startswith("\x1e"):
                    if buffer:
                        yield json.loads(buffer, object_hook=object_hook)
                    buffer = line.strip("\x1e")
                else:
                    buffer += line
            else:
                yield json.loads(buffer, object_hook=object_hook)

    else:

        def gen():
            yield json.loads(first_line, object_hook=object_hook)
            for line in lines:
                yield json.loads(line, object_hook=object_hook)

    return gen()


def nullable(val, cast):
    if val is None:
        return None
    else:
        return cast(val)


def eval_feature_expression(feature, expression):
    safe_dict = {"f": munchify(to_dict(feature))}
    safe_dict.update(
        {
            "sum": sum,
            "pow": pow,
            "min": min,
            "max": max,
            "math": math,
            "bool": bool,
            "int": partial(nullable, int),
            "str": partial(nullable, str),
            "float": partial(nullable, float),
            "len": partial(nullable, len),
        }
    )
    try:
        from shapely.geometry import shape

        safe_dict["shape"] = shape
    except ImportError:
        pass
    return eval(expression, {"__builtins__": None}, safe_dict)


def make_ld_context(context_items):
    """Returns a JSON-LD Context object.

    See https://json-ld.org/spec/latest/json-ld/."""
    ctx = {
        "@context": {
            "geojson": "http://ld.geojson.org/vocab#",
            "Feature": "geojson:Feature",
            "FeatureCollection": "geojson:FeatureCollection",
            "GeometryCollection": "geojson:GeometryCollection",
            "LineString": "geojson:LineString",
            "MultiLineString": "geojson:MultiLineString",
            "MultiPoint": "geojson:MultiPoint",
            "MultiPolygon": "geojson:MultiPolygon",
            "Point": "geojson:Point",
            "Polygon": "geojson:Polygon",
            "bbox": {"@container": "@list", "@id": "geojson:bbox"},
            "coordinates": "geojson:coordinates",
            "datetime": "http://www.w3.org/2006/time#inXSDDateTime",
            "description": "http://purl.org/dc/terms/description",
            "features": {"@container": "@set", "@id": "geojson:features"},
            "geometry": "geojson:geometry",
            "id": "@id",
            "properties": "geojson:properties",
            "start": "http://www.w3.org/2006/time#hasBeginning",
            "stop": "http://www.w3.org/2006/time#hasEnding",
            "title": "http://purl.org/dc/terms/title",
            "type": "@type",
            "when": "geojson:when",
        }
    }
    for item in context_items or []:
        t, uri = item.split("=")
        ctx[t.strip()] = uri.strip()
    return ctx


def id_record(rec):
    """Converts a record's id to a blank node id and returns the record."""
    rec["id"] = "_:f%s" % rec["id"]
    return rec


def recursive_round(obj, precision):
    """Recursively round coordinates."""
    if precision < 0:
        return obj
    if getattr(obj, "geometries", None):
        return Geometry(
            geometries=[recursive_round(part, precision) for part in obj.geometries]
        )
    elif getattr(obj, "coordinates", None):
        return Geometry(
            coordinates=[recursive_round(part, precision) for part in obj.coordinates]
        )
    if isinstance(obj, (int, float)):
        return round(obj, precision)
    else:
        return [recursive_round(part, precision) for part in obj]



"""
Support for TopoJSON was added in OGR 1.11 to the `GeoJSON` driver.
Starting at GDAL 2.3 support was moved to the `TopoJSON` driver.
"""

import os
import pytest

import fiona
from fiona.env import GDALVersion
from fiona.model import Properties


gdal_version = GDALVersion.runtime()

driver = "TopoJSON" if gdal_version.at_least((2, 3)) else "GeoJSON"
has_driver = driver in fiona.drvsupport.supported_drivers.keys()


@pytest.mark.skipif(not gdal_version.at_least((1, 11)), reason="Requires GDAL >= 1.11")
@pytest.mark.skipif(not has_driver, reason=f"Requires {driver} driver")
def test_read_topojson(data_dir):
    """Test reading a TopoJSON file

    The TopoJSON support in GDAL is a little unpredictable. In some versions
    the geometries or properties aren't parsed correctly. Here we just check
    that we can open the file, get the right number of features out, and
    that they have a geometry and some properties. See GH#722.
    """
    with fiona.open(os.path.join(data_dir, "example.topojson"), "r") as collection:
        features = list(collection)

    assert len(features) == 3, "unexpected number of features"
    for feature in features:
        assert isinstance(feature["properties"], Properties)
        assert len(feature["properties"]) > 0
        assert feature["geometry"]["type"] in {"Point", "LineString", "Polygon"}



"""Common commandline options for `fio`"""


from collections import defaultdict

import click


src_crs_opt = click.option('--src-crs', '--src_crs', help="Source CRS.")
dst_crs_opt = click.option('--dst-crs', '--dst_crs', help="Destination CRS.")


def cb_layer(ctx, param, value):
    """Let --layer be a name or index."""
    if value is None or not value.isdigit():
        return value
    else:
        return int(value)


def cb_multilayer(ctx, param, value):
    """
    Transform layer options from strings ("1:a,1:b", "2:a,2:c,2:z") to
    {
    '1': ['a', 'b'],
    '2': ['a', 'c', 'z']
    }
    """
    out = defaultdict(list)
    for raw in value:
        for v in raw.split(','):
            ds, name = v.split(':')
            out[ds].append(name)
    return out


def validate_multilayer_file_index(files, layerdict):
    """
    Ensure file indexes provided in the --layer option are valid
    """
    for key in layerdict.keys():
        if key not in [str(k) for k in range(1, len(files) + 1)]:
            layer = key + ":" + layerdict[key][0]
            raise click.BadParameter(f"Layer {layer} does not exist")



import fiona
import platform
import re
import os
import sys
from tests.conftest import travis_only
from fiona._env import GDALVersion, get_gdal_release_name


def test_version_tuple():
    version = fiona.gdal_version
    assert version.major >= 1 and isinstance(version.major, int)
    assert version.minor >= 0 and isinstance(version.minor, int)
    assert version.revision >= 0 and isinstance(version.revision, int)


def test_version_comparison():
    # version against version
    assert GDALVersion(4, 0, 0) > GDALVersion(3, 2, 1)
    assert GDALVersion(2, 0, 0) < GDALVersion(3, 2, 1)
    assert GDALVersion(3, 2, 2) > GDALVersion(3, 2, 1)
    assert GDALVersion(3, 2, 0) < GDALVersion(3, 2, 1)

    # tuple against version
    assert (4, 0, 0) > GDALVersion(3, 2, 1)
    assert (2, 0, 0) < GDALVersion(3, 2, 1)
    assert (3, 2, 2) > GDALVersion(3, 2, 1)
    assert (3, 2, 0) < GDALVersion(3, 2, 1)


@travis_only
def test_show_versions(capsys):
    version_pattern = re.compile(r"(\d+).(\d+).(\d+)")

    os_info = "{system} {release}".format(system=platform.system(),
                                          release=platform.release())
    python_version = platform.python_version()
    python_exec = sys.executable

    msg = ("Fiona version: {fiona_version}"
           "\nGDAL version: {gdal_release_name}"
           "\nPROJ version: {proj_version}"
           "\n"
           "\nOS: {os_info}"
           "\nPython: {python_version}"
           "\nPython executable: '{python_exec}'"
           "\n"
           )

    if fiona.gdal_version < GDALVersion(3, 0, 1):
        proj_version = "Proj version not available"
    else:
        proj_version = os.getenv("PROJVERSION")
        proj_version = re.match(version_pattern, proj_version).group(0)

    gdal_version = os.getenv("GDALVERSION")
    if not gdal_version == "master":
        gdal_version = re.match(version_pattern, gdal_version).group(0)
    else:
        gdal_version = get_gdal_release_name()

    msg_formatted = msg.format(fiona_version=fiona.__version__,
                               gdal_release_name=gdal_version,
                               proj_version=proj_version,
                               os_info=os_info,
                               python_version=python_version,
                               python_exec=python_exec)

    fiona.show_versions()
    captured = capsys.readouterr()

    assert captured.out.strip() == msg_formatted.strip()



"""Test listing a datasource's layers."""

import os

import pytest

import fiona
import fiona.ogrext
from fiona.errors import DriverError, FionaDeprecationWarning, FionaValueError
from fiona.io import ZipMemoryFile


def test_single_file_private(path_coutwildrnp_shp):
    with fiona.Env():
        assert fiona.ogrext._listlayers(
            path_coutwildrnp_shp) == ['coutwildrnp']


def test_single_file(path_coutwildrnp_shp):
    assert fiona.listlayers(path_coutwildrnp_shp) == ['coutwildrnp']


def test_directory(data_dir):
    assert sorted(fiona.listlayers(data_dir)) == ['coutwildrnp', 'gre', 'test_tin']


def test_directory_trailing_slash(data_dir):
    assert sorted(fiona.listlayers(data_dir)) == ['coutwildrnp', 'gre', 'test_tin']


def test_zip_path(path_coutwildrnp_zip):
    assert fiona.listlayers(
        f'zip://{path_coutwildrnp_zip}') == ['coutwildrnp']


def test_zip_path_arch(path_coutwildrnp_zip):
    vfs = f'zip://{path_coutwildrnp_zip}'
    with pytest.warns(FionaDeprecationWarning):
        assert fiona.listlayers('/coutwildrnp.shp', vfs=vfs) == ['coutwildrnp']


def test_list_not_existing(data_dir):
    """Test underlying Cython function correctly raises"""
    path = os.path.join(data_dir, "does_not_exist.geojson")
    with pytest.raises(DriverError):
        fiona.ogrext._listlayers(path)


def test_invalid_path():
    with pytest.raises(TypeError):
        fiona.listlayers(1)


def test_invalid_vfs():
    with pytest.raises(TypeError):
        fiona.listlayers("/", vfs=1)


def test_invalid_path_ioerror():
    with pytest.raises(DriverError):
        fiona.listlayers("foobar")


def test_listing_file(path_coutwildrnp_json):
    """list layers from an open file object"""
    with open(path_coutwildrnp_json, "rb") as f:
        assert len(fiona.listlayers(f)) == 1


def test_listing_pathobj(path_coutwildrnp_json):
    """list layers from a Path object"""
    pathlib = pytest.importorskip("pathlib")
    assert len(fiona.listlayers(pathlib.Path(path_coutwildrnp_json))) == 1


def test_listdir_path(path_coutwildrnp_zip):
    """List directories in a path"""
    assert sorted(fiona.listdir("zip://{}".format(path_coutwildrnp_zip))) == [
        "coutwildrnp.dbf",
        "coutwildrnp.prj",
        "coutwildrnp.shp",
        "coutwildrnp.shx",
    ]


def test_listdir_path_not_existing(data_dir):
    """Test listing of a non existent directory"""
    path = os.path.join(data_dir, "does_not_exist.zip")
    with pytest.raises(FionaValueError):
        fiona.listdir(path)


def test_listdir_invalid_path():
    """List directories with invalid path"""
    with pytest.raises(TypeError):
        assert fiona.listdir(1)


def test_listdir_file(path_coutwildrnp_zip):
    """Test list directories of a file"""
    with pytest.raises(FionaValueError):
        fiona.listdir(f"zip://{path_coutwildrnp_zip}/coutwildrnp.shp")


def test_listdir_zipmemoryfile(bytes_coutwildrnp_zip):
    """Test list directories of a zipped memory file."""
    with ZipMemoryFile(bytes_coutwildrnp_zip) as memfile:
        print(memfile.name)
        assert sorted(fiona.listdir(memfile.name)) == [
            "coutwildrnp.dbf",
            "coutwildrnp.prj",
            "coutwildrnp.shp",
            "coutwildrnp.shx",
        ]



import pytest

import fiona
from fiona.collection import supported_drivers
from fiona.errors import FionaValueError, DriverError, SchemaError, CRSError
from fiona.model import Feature


def test_json_read(path_coutwildrnp_json):
    with fiona.open(path_coutwildrnp_json, "r") as c:
        assert len(c) == 67


def test_json(tmpdir):
    """Write a simple GeoJSON file"""
    path = str(tmpdir.join("foo.json"))
    with fiona.open(
        path,
        "w",
        driver="GeoJSON",
        schema={"geometry": "Unknown", "properties": [("title", "str")]},
    ) as c:
        c.writerecords(
            [
                Feature.from_dict(
                    **{
                        "geometry": {"type": "Point", "coordinates": [0.0, 0.0]},
                        "properties": {"title": "One"},
                    }
                )
            ]
        )
        c.writerecords(
            [
                Feature.from_dict(
                    **{
                        "geometry": {"type": "MultiPoint", "coordinates": [[0.0, 0.0]]},
                        "properties": {"title": "Two"},
                    }
                )
            ]
        )
    with fiona.open(path) as c:
        assert c.schema["geometry"] == "Unknown"
        assert len(c) == 2


def test_json_overwrite(tmpdir):
    """Overwrite an existing GeoJSON file"""
    path = str(tmpdir.join("foo.json"))

    driver = "GeoJSON"
    schema1 = {"geometry": "Unknown", "properties": [("title", "str")]}
    schema2 = {"geometry": "Unknown", "properties": [("other", "str")]}

    features1 = [
        Feature.from_dict(
            **{
                "geometry": {"type": "Point", "coordinates": [0.0, 0.0]},
                "properties": {"title": "One"},
            }
        ),
        Feature.from_dict(
            **{
                "geometry": {"type": "MultiPoint", "coordinates": [[0.0, 0.0]]},
                "properties": {"title": "Two"},
            }
        ),
    ]
    features2 = [
        Feature.from_dict(
            **{
                "geometry": {"type": "Point", "coordinates": [0.0, 0.0]},
                "properties": {"other": "Three"},
            }
        ),
    ]

    # write some data to a file
    with fiona.open(path, "w", driver=driver, schema=schema1) as c:
        c.writerecords(features1)

    # test the data was written correctly
    with fiona.open(path, "r") as c:
        assert len(c) == 2
        feature = next(iter(c))
        assert feature.properties["title"] == "One"

    # attempt to overwrite the existing file with some new data
    with fiona.open(path, "w", driver=driver, schema=schema2) as c:
        c.writerecords(features2)

    # test the second file was written correctly
    with fiona.open(path, "r") as c:
        assert len(c) == 1
        feature = next(iter(c))
        assert feature.properties["other"] == "Three"


def test_json_overwrite_invalid(tmpdir):
    """Overwrite an existing file that isn't a valid GeoJSON"""

    # write some invalid data to a file
    path = str(tmpdir.join("foo.json"))
    with open(path, "w") as f:
        f.write("This isn't a valid GeoJSON file!!!")

    schema1 = {"geometry": "Unknown", "properties": [("title", "str")]}
    features1 = [
        Feature.from_dict(
            **{
                "geometry": {"type": "Point", "coordinates": [0.0, 0.0]},
                "properties": {"title": "One"},
            }
        ),
        Feature.from_dict(
            **{
                "geometry": {"type": "MultiPoint", "coordinates": [[0.0, 0.0]]},
                "properties": {"title": "Two"},
            }
        ),
    ]

    # attempt to overwrite it with a valid file
    with fiona.open(path, "w", driver="GeoJSON", schema=schema1) as dst:
        dst.writerecords(features1)

    # test the data was written correctly
    with fiona.open(path, "r") as src:
        assert len(src) == 2


def test_write_json_invalid_directory(tmpdir):
    """Attempt to create a file in a directory that doesn't exist"""
    path = str(tmpdir.join("does-not-exist", "foo.json"))
    schema = {"geometry": "Unknown", "properties": [("title", "str")]}
    with pytest.raises(DriverError):
        fiona.open(path, "w", driver="GeoJSON", schema=schema)



"""$ fio load"""

from functools import partial

import click
import cligj

import fiona
from fiona.fio import options, with_context_env
from fiona.model import Feature, Geometry
from fiona.schema import FIELD_TYPES_MAP_REV
from fiona.transform import transform_geom


def _cb_key_val(ctx, param, value):
    """
    click callback to validate `--opt KEY1=VAL1 --opt KEY2=VAL2` and collect
    in a dictionary like the one below, which is what the CLI function receives.
    If no value or `None` is received then an empty dictionary is returned.

        {
            'KEY1': 'VAL1',
            'KEY2': 'VAL2'
        }

    Note: `==VAL` breaks this as `str.split('=', 1)` is used.

    """
    if not value:
        return {}
    else:
        out = {}
        for pair in value:
            if "=" not in pair:
                raise click.BadParameter(
                    f"Invalid syntax for KEY=VAL arg: {pair}"
                )
            else:
                k, v = pair.split("=", 1)
                k = k.lower()
                v = v.lower()
                out[k] = None if v.lower() in ["none", "null", "nil", "nada"] else v
        return out


@click.command(short_help="Load GeoJSON to a dataset in another format.")
@click.argument("output", required=True)
@click.option("-f", "--format", "--driver", "driver", help="Output format driver name.")
@options.src_crs_opt
@click.option(
    "--dst-crs",
    "--dst_crs",
    help="Destination CRS.  Defaults to --src-crs when not given.",
)
@cligj.features_in_arg
@click.option(
    "--layer",
    metavar="INDEX|NAME",
    callback=options.cb_layer,
    help="Load features into specified layer.  Layers use "
    "zero-based numbering when accessed by index.",
)
@click.option(
    "--co",
    "--profile",
    "creation_options",
    metavar="NAME=VALUE",
    multiple=True,
    callback=_cb_key_val,
    help="Driver specific creation options. See the documentation for the selected output driver for more information.",
)
@click.pass_context
@with_context_env
def load(ctx, output, driver, src_crs, dst_crs, features, layer, creation_options):
    """Load features from JSON to a file in another format.

    The input is a GeoJSON feature collection or optionally a sequence of
    GeoJSON feature objects.

    """
    dst_crs = dst_crs or src_crs

    if src_crs and dst_crs and src_crs != dst_crs:
        transformer = partial(
            transform_geom, src_crs, dst_crs, antimeridian_cutting=True
        )
    else:

        def transformer(x):
            return Geometry.from_dict(**x)

    def feature_gen():
        """Convert stream of JSON to features.

        Yields
        ------
        Feature

        """
        try:
            for feat in features:
                feat["geometry"] = transformer(Geometry.from_dict(**feat["geometry"]))
                yield Feature.from_dict(**feat)
        except TypeError:
            raise click.ClickException("Invalid input.")

    source = feature_gen()

    # Use schema of first feature as a template.
    # TODO: schema specified on command line?
    try:
        first = next(source)
    except TypeError:
        raise click.ClickException("Invalid input.")

    # print(first, first.geometry)
    schema = {"geometry": first.geometry.type}
    schema["properties"] = {
            k: FIELD_TYPES_MAP_REV.get(type(v)) or "str"
            for k, v in first.properties.items()
    }

    with fiona.open(
        output,
        "w",
        driver=driver,
        crs=dst_crs,
        schema=schema,
        layer=layer,
        **creation_options
    ) as dst:
        dst.write(first)
        dst.writerecords(source)
